from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Dict, Union, List

from unittest import TestCase
from unittest.mock import patch

from obspy.core import UTCDateTime
import numpy as np

import sohstationviewer.view.plotting.time_power_squared_processor
from sohstationviewer.conf import constants as const
from sohstationviewer.model.handling_data import (
    trim_downsample_soh_chan,
    trim_downsample_wf_chan,
    trim_waveform_data,
    downsample_waveform_data,
    get_each_day_5_min_list,
)
from sohstationviewer.view.plotting.time_power_squared_processor import (
    TimePowerSquaredProcessor,
)
from sohstationviewer.model.downsampler import downsample, chunk_minmax

ORIGINAL_CHAN_SIZE_LIMIT = const.CHAN_SIZE_LIMIT
ORIGINAL_RECAL_SIZE_LIMIT = const.RECAL_SIZE_LIMIT
ZERO_EPOCH_TIME = UTCDateTime(1970, 1, 1, 0, 0, 0).timestamp
TraceInfo = Dict[str, Union[float, str]]
ChannelData = Dict[str,
                   Union[float, List[np.ndarray], List[TraceInfo], Dict,
                         np.ndarray]]


class TestTrimWfData(TestCase):
    def setUp(self) -> None:
        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info

        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            self.traces_info.append(trace)
        self.start_time = 2500
        self.end_time = 7500

    def test_data_is_trimmed_neither_start_nor_end_time_is_trace_start_or_end_time(self):  # noqa: E501
        self.start_time = 2444
        self.end_time = 7444
        trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )

        self.assertTrue(
            trimmed_traces_list[0]['startTmEpoch'] <= self.start_time)
        self.assertTrue(
            trimmed_traces_list[0]['endTmEpoch'] > self.start_time
        )
        trimmed_traces_list.pop(0)
        trimmed_traces_list.pop()
        is_left_trimmed = all(trace['startTmEpoch'] > self.start_time
                              for trace in trimmed_traces_list)
        is_right_trimmed = all(trace['endTmEpoch'] <= self.end_time
                               for trace in trimmed_traces_list)
        self.assertTrue(is_left_trimmed and is_right_trimmed)

    def test_data_out_of_range(self):
        with self.subTest('test_start_time_later_than_data_end_time'):
            self.start_time = 12500
            self.end_time = 17500
            self.assertFalse(
                trim_downsample_wf_chan(self.channel_data, self.start_time,
                                        self.end_time, True)
            )
        with self.subTest('test_end_time_earlier_than_data_start_time'):
            self.start_time = -7500
            self.end_time = -2500
            self.assertFalse(
                trim_downsample_wf_chan(self.channel_data, self.start_time,
                                        self.end_time, True)
            )

    def test_no_data(self):
        self.channel_data['tracesInfo'] = []
        with self.assertRaises(IndexError):
            trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )

    def test_end_time_earlier_than_start_time(self):
        self.start_time, self.end_time = self.end_time, self.start_time
        trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )
        self.assertListEqual(trimmed_traces_list, [])

    def test_data_does_not_need_to_be_trimmed(self):
        with self.subTest('test_start_time_earlier_than_trace_earliest_time'):
            self.start_time = -2500
            self.end_time = 7500
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), 76)
        with self.subTest('test_end_time_later_than_trace_latest_time'):
            self.start_time = 2500
            self.end_time = 12500
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), 75)
        with self.subTest('test_data_contained_in_time_range'):
            self.start_time = self.traces_info[0]['startTmEpoch']
            self.end_time = self.traces_info[-1]['endTmEpoch']
            trimmed_traces_list = trim_waveform_data(
                self.channel_data, self.start_time, self.end_time
            )
            self.assertEqual(len(trimmed_traces_list), len(self.traces_info))


class TestDownsampleWaveformData(TestCase):
    def no_file_memmap(self, file_path: Path, **kwargs):
        # Data will look the same as times. This has two benefits:
        # - It is a lot easier to inspect what data remains after trimming
        # and downsampling, seeing as the remaining data would be the same
        # as the remaining times.
        # - It is a lot easier to reproducibly create a test data set.
        array_size = 100
        file_idx = int(file_path.name.split('-')[-1])
        start = file_idx * array_size
        end = start + array_size
        return np.arange(start, end)

    def setUp(self) -> None:
        memmap_patcher = patch.object(np, 'memmap',
                                      side_effect=self.no_file_memmap)
        self.addCleanup(memmap_patcher.stop)
        memmap_patcher.start()

        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info
        self.data_folder = TemporaryDirectory()
        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            trace['size'] = trace_size

            times_file_name = Path(self.data_folder.name) / f'times-{i}'
            trace['times_f'] = times_file_name

            data_file_name = Path(self.data_folder.name) / f'data-{i}'
            trace['data_f'] = data_file_name

            self.traces_info.append(trace)
        self.start_time = 2550
        self.end_time = 7550
        self.trimmed_traces_list = trim_waveform_data(
            self.channel_data, self.start_time, self.end_time
        )

    @patch('sohstationviewer.model.handling_data.downsample', wraps=downsample)
    def test_data_is_downsampled(self, mock_downsample):
        const.CHAN_SIZE_LIMIT = 1000
        downsample_waveform_data(self.trimmed_traces_list,
                                 self.start_time, self.end_time)
        self.assertTrue(mock_downsample.called)
        const.CHAN_SIZE_LIMIT = ORIGINAL_CHAN_SIZE_LIMIT

    def test_all_traces_handled(self):
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertEqual(len(downsampled_times), 51)
        self.assertEqual(len(downsampled_data), 51)

    def test_downsampling_not_needed(self):
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        with self.subTest('test_data_points_outside_time_range_removed'):
            self.assertEqual(downsampled_times.pop(0).size, 50)
            self.assertEqual(downsampled_times.pop(-1).size, 51)
            self.assertEqual(downsampled_data.pop(0).size, 50)
            self.assertEqual(downsampled_data.pop(-1).size, 51)

        with self.subTest('test_intermediate_data_points_not_removed'):
            self.assertTrue(
                all(times.size == 100 for times in downsampled_times)
            )
            self.assertTrue(
                all(data.size == 100 for data in downsampled_times)
            )

    def test_trace_list_empty(self):
        self.trimmed_traces_list = []
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertListEqual(downsampled_times, [])
        self.assertListEqual(downsampled_data, [])

    def test_end_time_earlier_than_start_time(self):
        self.start_time, self.end_time = self.end_time, self.start_time
        downsampled_times, downsampled_data = downsample_waveform_data(
            self.trimmed_traces_list,
            self.start_time, self.end_time
        )
        self.assertTrue(all(times.size == 0 for times in downsampled_times))
        self.assertTrue(all(data.size == 0 for data in downsampled_data))

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        const.CHAN_SIZE_LIMIT = 1000
        downsample_waveform_data(self.trimmed_traces_list,
                                 self.start_time, self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 2)


class TestDownsample(TestCase):
    def setUp(self) -> None:
        patcher = patch('sohstationviewer.model.downsampler.chunk_minmax')
        self.addCleanup(patcher.stop)
        self.mock_chunk_minmax = patcher.start()
        self.times = np.arange(1000)
        self.data = np.arange(1000)
        self.log_idx = np.arange(1000)

    def test_first_downsample_step_remove_enough_points(self):
        req_points = 999
        downsample(self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)

    def test_first_downsample_step_remove_enough_points_with_logidx(self):
        req_points = 999
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)

    def test_second_downsample_step_required(self):
        req_points = 1
        downsample(self.times, self.data, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, _, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertEqual(rq_points, req_points)

    def test_second_downsample_step_required_with_logidx(self):
        req_points = 1
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, log_idx, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertIsNot(log_idx, self.log_idx)
        self.assertEqual(rq_points, req_points)

    def test_requested_points_greater_than_data_size(self):
        req_points = 10000
        times, data, _ = downsample(
            self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)

    def test_requested_points_greater_than_data_size_with_logidx(self):
        req_points = 10000
        times, data, log_idx = downsample(
            self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)
        self.assertIs(log_idx, self.log_idx)

    def test_requested_points_is_zero(self):
        req_points = 0
        downsample(self.times, self.data, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, _, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertEqual(rq_points, req_points)

    def test_requested_points_is_zero_with_logidx(self):
        req_points = 0
        downsample(self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertTrue(self.mock_chunk_minmax.called)
        times, data, log_idx, rq_points = self.mock_chunk_minmax.call_args[0]
        self.assertIsNot(times, self.times)
        self.assertIsNot(data, self.data)
        self.assertIsNot(log_idx, self.log_idx)
        self.assertEqual(rq_points, req_points)

    def test_empty_times_and_data(self):
        req_points = 1000
        self.times = np.empty((0, 0))
        self.data = np.empty((0, 0))
        times, data, _ = downsample(
            self.times, self.data, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)

    def test_empty_times_and_data_with_logidx(self):
        req_points = 1000
        self.times = np.empty((0, 0))
        self.data = np.empty((0, 0))
        self.log_idx = np.empty((0, 0))
        times, data, log_idx = downsample(
            self.times, self.data, self.log_idx, rq_points=req_points)
        self.assertFalse(self.mock_chunk_minmax.called)
        # Check that we did not do any processing on the times and data arrays.
        # This ensures that we don't do two unneeded copy operations.
        self.assertIs(times, self.times)
        self.assertIs(data, self.data)
        self.assertIs(log_idx, self.log_idx)


class TestChunkMinmax(TestCase):
    def setUp(self):
        self.times = np.arange(1000)
        self.data = np.arange(1000)
        self.log_idx = np.arange(1000)

    def test_data_size_is_multiple_of_requested_points(self):
        req_points = 100
        times, data, log_idx = chunk_minmax(
            self.times, self.data, self.log_idx, req_points)
        self.assertEqual(times.size, req_points)
        self.assertEqual(data.size, req_points)
        self.assertEqual(log_idx.size, req_points)

    @patch('sohstationviewer.model.downsampler.downsample', wraps=downsample)
    def test_data_size_is_not_multiple_of_requested_points(
            self, mock_downsample):
        req_points = 102
        chunk_minmax(self.times, self.data, self.log_idx, req_points)
        self.assertTrue(mock_downsample.called)

    def test_requested_points_too_small(self):
        small_req_points_list = [0, 1]
        for req_points in small_req_points_list:
            with self.subTest(f'test_requested_points_is_{req_points}'):
                times, data, log_idx = chunk_minmax(
                    self.times, self.data, self.log_idx, rq_points=req_points)
                self.assertEqual(times.size, 0)
                self.assertEqual(data.size, 0)
                self.assertEqual(data.size, 0)


class TestTrimDownsampleSohChan(TestCase):
    @staticmethod
    def downsample(times, data, log_indexes=None, rq_points=0):
        return times, data, log_indexes

    def setUp(self) -> None:
        self.channel_info = {}
        self.traces_info = [{
            'times': np.arange(1000),
            'data': np.arange(1000)
        }]
        self.channel_info['tracesInfo'] = self.traces_info
        self.start_time = 250
        self.end_time = 750
        self.first_time = False

        patcher = patch('sohstationviewer.model.handling_data.downsample')
        self.addCleanup(patcher.stop)
        self.mock_downsample = patcher.start()
        self.mock_downsample.side_effect = self.downsample

    def num_points_outside_time_range(self, start_time, end_time):
        return len([data_point
                    for data_point in self.traces_info[0]['times']
                    if not start_time <= data_point <= end_time])

    def test_start_time_later_than_times_data(self):
        self.start_time = 250
        self.end_time = 1250
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertGreaterEqual(self.channel_info['times'][0].min(),
                                self.start_time)
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_end_time_earlier_than_times_data(self):
        self.start_time = -250
        self.end_time = 750
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertLessEqual(self.channel_info['times'][0].max(),
                             self.end_time)
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_start_time_earlier_than_times_data(self):
        self.start_time = -250
        self.end_time = 750
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertEqual(self.channel_info['times'][0].min(),
                         self.channel_info['tracesInfo'][0]['times'].min())
        self.assertEqual(
            self.traces_info[0]['times'].size
            - self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_end_time_later_than_times_data(self):
        self.start_time = 250
        self.end_time = 1250
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertEqual(self.channel_info['times'][0].max(),
                         self.channel_info['tracesInfo'][0]['times'].max())
        self.assertEqual(
            self.traces_info[0]['times'].size -
            self.channel_info['times'][0].size,
            self.num_points_outside_time_range(self.start_time, self.end_time)
        )

    def test_times_data_contained_in_time_range(self):
        self.start_time = -250
        self.end_time = 1250
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        np.testing.assert_array_equal(self.channel_info['times'][0],
                                      self.traces_info[0]['times'])

    def test_time_range_is_the_same_as_times_data(self):
        self.start_time = ZERO_EPOCH_TIME
        self.end_time = 999
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        np.testing.assert_array_equal(self.channel_info['times'][0],
                                      self.traces_info[0]['times'])

    def test_time_range_does_not_overlap_times_data(self):
        self.start_time = 2000
        self.end_time = 3000
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertEqual(len(self.channel_info['times']), 0)
        self.assertEqual(len(self.channel_info['data']), 0)

    def test_data_is_downsampled(self):
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertTrue(self.mock_downsample.called)

    def test_processed_data_is_stored_in_appropriate_location(self):
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        expected_keys = ('tracesInfo', 'times', 'data')
        self.assertTupleEqual(tuple(self.channel_info.keys()),
                              expected_keys)

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 2)


class TestTrimDownsampleSohChanWithLogidx(TestCase):
    def setUp(self) -> None:
        self.channel_info = {'reftek': True}
        self.traces_info = [{
            'times': np.arange(1000),
            'data': np.arange(1000),
            'logIdx': np.arange(1000)
        }]
        self.channel_info['tracesInfo'] = self.traces_info
        self.start_time = 250
        self.end_time = 750
        self.first_time = False

    def test_time_range_does_not_overlap_times_data(self):
        self.start_time = 2000
        self.end_time = 3000
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        self.assertEqual(len(self.channel_info['times']), 0)
        self.assertEqual(len(self.channel_info['data']), 0)
        self.assertEqual(len(self.channel_info['logIdx']), 0)

    def test_processed_data_is_stored_in_appropriate_location(self):
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)
        expected_keys = ('reftek', 'tracesInfo', 'times', 'data', 'logIdx')
        self.assertTupleEqual(tuple(self.channel_info.keys()),
                              expected_keys)

    @patch('sohstationviewer.model.handling_data.downsample')
    def test_arguments_sent_to_downsample(self, mock_downsample):
        mock_downsample.return_value = (1, 2, 3)
        trim_downsample_soh_chan(self.channel_info, self.start_time,
                                 self.end_time)

        positional_args, named_args = mock_downsample.call_args
        self.assertEqual(len(positional_args), 3)


class TestTrimDownsampleWfChan(TestCase):
    def no_file_memmap(self, file_path: Path, **kwargs):
        # Data will look the same as times. This has two benefits:
        # - It is a lot easier to inspect what data remains after trimming
        # and downsampling, seeing as the remaining data would be the same
        # as the remaining times.
        # - It is a lot easier to reproducibly create a test data set.
        array_size = 100
        file_idx = int(file_path.name.split('-')[-1])
        start = file_idx * array_size
        end = start + array_size
        return np.arange(start, end)

    def setUp(self) -> None:
        memmap_patcher = patch.object(np, 'memmap',
                                      side_effect=self.no_file_memmap)
        self.addCleanup(memmap_patcher.stop)
        memmap_patcher.start()

        self.channel_data = {}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info
        self.data_folder = TemporaryDirectory()
        for i in range(100):
            trace_size = 100
            start_time = i * trace_size
            trace = {}
            trace['startTmEpoch'] = start_time
            trace['endTmEpoch'] = start_time + trace_size - 1
            trace['size'] = trace_size

            times_file_name = Path(self.data_folder.name) / f'times-{i}'
            trace['times_f'] = times_file_name

            data_file_name = Path(self.data_folder.name) / f'data-{i}'
            trace['data_f'] = data_file_name

            self.traces_info.append(trace)
        self.start_time = 2500
        self.end_time = 7500

    def test_result_is_stored(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertTrue('times' in self.channel_data)
        self.assertGreater(len(self.channel_data['times']), 0)
        self.assertTrue('data' in self.channel_data)
        self.assertGreater(len(self.channel_data['data']), 0)

    def test_data_small_enough_after_first_trim_flag_is_set(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertTrue('fulldata' in self.channel_data)

    def test_no_additional_work_if_data_small_enough_after_first_trim(self):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        current_times = self.channel_data['times']
        current_data = self.channel_data['data']
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, True)
        self.assertIs(current_times, self.channel_data['times'])
        self.assertIs(current_data, self.channel_data['data'])

    def test_data_too_large_after_trimming(self):
        const.RECAL_SIZE_LIMIT = 1
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, False)
        self.assertTrue('times' not in self.channel_data)
        self.assertTrue('data' not in self.channel_data)
        const.RECAL_SIZE_LIMIT = ORIGINAL_RECAL_SIZE_LIMIT

    @patch('sohstationviewer.model.handling_data.trim_waveform_data',
           wraps=trim_waveform_data)
    @patch('sohstationviewer.model.handling_data.downsample_waveform_data',
           wraps=downsample_waveform_data)
    def test_data_trim_and_downsampled(self, mock_downsample, mock_trim):
        trim_downsample_wf_chan(self.channel_data, self.start_time,
                                self.end_time, False)
        self.assertTrue(mock_trim.called)
        self.assertTrue(mock_downsample.called)


class TestGetTrimTpsData(TestCase):
    def no_file_memmap(self, file_path: Path, *args, **kwargs):
        """
        A mock of numpy.memmap. Reduce test run time significantly by making
        sure that data access happens in memory and not on disk.

        This method does not actually load the data stored on disk. Instead, it
        constructs the array of data using the name of the given file. To do
        so, this method requires the file name to be in the format
        <prefix>_<index>. This method then constructs an array of
        self.trace_size consecutive integers starting at
        <index> * self.trace_size.

        :param file_path: the path to a file used to construct the data array.
        :param args: dummy arguments to make the API similar to numpy.memmap.
        :param kwargs: dummy arguments to make the API similar to numpy.memmap.
        :return: a numpy array constructed using file_path's name.
        """
        file_idx = int(file_path.name.split('_')[-1])
        start = file_idx * self.trace_size
        end = start + self.trace_size
        return np.arange(start, end)

    def add_trace(self, start_time: float, idx: Optional[int] = None):
        """
        Add a trace to the stored list of traces.

        :param start_time: the start time of the trace to be added.
        :param idx: the index to insert the trace into. If None, the new trace
            will be appended to the list of traces
        """
        trace = {}
        trace['startTmEpoch'] = start_time
        trace['endTmEpoch'] = start_time + self.trace_size - 1
        trace['size'] = self.trace_size

        file_idx = start_time // self.trace_size
        times_file_name = Path(self.data_folder.name) / f'times_{file_idx}'
        trace['times_f'] = times_file_name

        data_file_name = Path(self.data_folder.name) / f'data_{file_idx}'
        trace['data_f'] = data_file_name

        if idx is not None:
            self.traces_info.insert(idx, trace)
        else:
            self.traces_info.append(trace)

    def setUp(self) -> None:
        """Set up text fixtures."""
        memmap_patcher = patch.object(np, 'memmap',
                                      side_effect=self.no_file_memmap)
        self.addCleanup(memmap_patcher.stop)
        memmap_patcher.start()

        # Channel ID is only used when communicating with the main window.
        # Seeing as we are testing the processing step here, we don't really
        # need it.
        channel_id = ''

        self.channel_data: ChannelData = {'samplerate': 1}
        self.traces_info = []
        self.channel_data['tracesInfo'] = self.traces_info
        self.data_folder = TemporaryDirectory()
        self.trace_size = 1000
        for i in range(100):
            start_time = i * self.trace_size
            self.add_trace(start_time)
        self.start_time = 25000
        self.end_time = 75000
        self.each_day_5_mins_list = get_each_day_5_min_list(self.start_time,
                                                            self.end_time)
        self.tps_processor = TimePowerSquaredProcessor(
            channel_id, self.channel_data, self.start_time, self.end_time,
            self.each_day_5_mins_list
        )

    local_TimePowerSquaredProcessor = (sohstationviewer.view.plotting.
                                       time_power_squared_processor.
                                       TimePowerSquaredProcessor)

    # If object obj is instance of class A, then the method call obj.method1()
    # translate to A.method1(obj) for Python. So, in order to mock method1 for
    # obj, we mock it for the class A.
    @patch.object(local_TimePowerSquaredProcessor, 'trim_waveform_data')
    def test_data_is_trimmed(self, mock_trim_waveform_data):
        """Test that the data is trimmed."""
        self.tps_processor.run()
        self.assertTrue(mock_trim_waveform_data.called)

    def test_appropriate_amount_of_5_mins_skipped(self):
        """Test that the trimmed part of the data is skipped over."""
        self.tps_processor.run()
        with self.subTest('test_skip_before_start_time'):
            first_unskipped_idx = 83
            skipped_tps_arr = (
                self.channel_data['tps_data'][0][:first_unskipped_idx]
            )
            self.assertTrue((skipped_tps_arr == 0).all())
        with self.subTest('test_skip_after_end_time'):
            last_unskipped_idx = 252
            skipped_tps_arr = (
                self.channel_data['tps_data'][0][last_unskipped_idx + 1:]
            )
            self.assertTrue((skipped_tps_arr == 0).all())

    def test_result_is_stored(self):
        """Test that the result of the TPS calculation is stored."""
        self.tps_processor.run()
        self.assertTrue('tps_data' in self.channel_data)

    def test_formula_is_correct(self):
        """Test that the TPS calculation uses the correct formula."""
        self.tps_processor.start_time = 50000
        self.tps_processor.end_time = 52000
        self.tps_processor.run()
        first_unskipped_idx = 166
        last_unskipped_idx = 175
        tps_data = self.channel_data['tps_data'][0]
        unskipped_tps_arr = (
            tps_data[first_unskipped_idx:last_unskipped_idx + 1]
        )
        expected = np.array([
            2.51497985e+09, 2.54515955e+09, 2.57551925e+09, 0.00000000e+00,
            1.96222188e+09, 2.64705855e+09, 2.67801825e+09, 2.03969638e+09,
            2.75095755e+09, 2.78251725e+09
        ])
        self.assertTrue(np.allclose(unskipped_tps_arr, expected))

    def test_one_tps_array_for_each_day_one_day_of_data(self):
        """
        Test that there is one TPS array for each day of data.

        Test the case where there is only one day of data.
        """
        self.tps_processor.run()
        self.assertEqual(len(self.channel_data['tps_data']), 1)

    def test_one_tps_array_for_each_day_multiple_days_of_data(self):
        """
        Test that there is one TPS array for each dat of data.

        Test the case where there are more than one day of data.
        """
        # Currently, the data time goes from 0 to 100000, which is enough to
        # cover two days (the start of the second positive day in epoch time is
        # 86400). Thus, we only have to set the end time to the data end time
        # to have two days of data.
        self.tps_processor.end_time = 100000
        self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
            self.tps_processor.start_time, self.tps_processor.end_time
        )
        self.tps_processor.run()
        self.assertEqual(len(self.channel_data['tps_data']), 2)

    def test_data_has_gap_to_the_right_data_same_day_before_gap(self):
        """
        Test that gaps in the data are skipped in TPS calculation by checking
        that the elements in the TPS array corresponding to the gaps are
        0.

        Test the case where there are gaps to the right of the data and the
        traces directly next to the gaps are in the same day.
        """
        # Remove traces that go from 1000 to 24999 (traces 2 to 25) in order to
        # create a gap on the right side of the data.
        self.traces_info = [trace
                            for i, trace in enumerate(self.traces_info)
                            if not 0 < i < 25]
        self.channel_data['tracesInfo'] = self.traces_info

        with self.subTest('test_start_time_in_gap'):
            self.tps_processor.start_time = 15000
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 1)
            tps_gap = slice(0, 50)
            tps_data_in_gap = self.channel_data['tps_data'][0][tps_gap]
            tps_data_in_gap_contains_zero = np.allclose(
                tps_data_in_gap, np.zeros(tps_data_in_gap.size)
            )
            self.assertTrue(tps_data_in_gap_contains_zero)

        with self.subTest('test_start_time_cover_all_traces'):
            self.tps_processor.start_time = 500
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 1)
            tps_gap = slice(2, 83)
            tps_data_in_gap = self.channel_data['tps_data'][0][tps_gap]
            tps_data_in_gap_contains_zero = np.allclose(
                tps_data_in_gap, np.zeros(tps_data_in_gap.size)
            )
            self.assertTrue(tps_data_in_gap_contains_zero)

    def test_data_has_gap_to_the_left_data_same_day_after_gap(self):
        """
        Test that gaps in the data are skipped in TPS calculation by checking
        that the elements in the TPS array corresponding to the gaps are
        0.

        Test the case where there are gaps to the left of the data and the
        traces directly next to the gaps are in the same day.
        """
        # Data end time is 100000, so we want a trace that starts after 100001
        trace_start_time = 125000
        self.add_trace(trace_start_time)

        with self.subTest('test_end_time_in_gap'):
            # Subject to change after Issue #37 is fixed
            self.tps_processor.end_time = 110000
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 2)
            tps_gaps = (slice(45, 128), slice(131, None))
            tps_data_in_gaps = np.concatenate(
                [self.channel_data['tps_data'][1][gap] for gap in tps_gaps]
            )
            tps_data_in_gaps_contains_zero = np.allclose(
                tps_data_in_gaps, np.zeros(tps_data_in_gaps.size)
            )
            self.assertTrue(tps_data_in_gaps_contains_zero)

        with self.subTest('test_end_time_cover_all_traces'):
            self.tps_processor.end_time = trace_start_time + 50
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 2)
            tps_gaps = (slice(45, 128), slice(131, None))
            tps_data_in_gaps = np.concatenate(
                [self.channel_data['tps_data'][1][gap] for gap in tps_gaps]
            )
            tps_data_in_gaps_contains_zero = np.allclose(
                tps_data_in_gaps, np.zeros(tps_data_in_gaps.size)
            )
            self.assertTrue(tps_data_in_gaps_contains_zero)

    def test_data_has_gap_to_the_right_data_different_day_before_gap(self):
        """
        Test that gaps in the data are skipped in TPS calculation by checking
        that the elements in the TPS array corresponding to the gaps are
        0.

        Test the case where there are gaps to the right of the data and the
        traces directly next to the gaps are in different days.
        """
        trace_start_time = -50000
        self.add_trace(trace_start_time, idx=0)

        with self.subTest('test_start_time_in_gap'):
            self.tps_processor.start_time = -25000
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 2)
            tps_gap = slice(const.NO_5M_DAY)
            tps_data_in_gap = self.channel_data['tps_data'][0][tps_gap]
            tps_data_in_gap_contains_zero = np.allclose(
                tps_data_in_gap, np.zeros(tps_data_in_gap.size)
            )
            self.assertTrue(tps_data_in_gap_contains_zero)

        with self.subTest('test_start_time_cover_all_traces'):
            self.tps_processor.start_time = -60000
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 2)
            tps_gaps = (slice(0, 121), slice(124, None))
            tps_data_in_gaps = np.concatenate(
                [self.channel_data['tps_data'][0][gap] for gap in tps_gaps]
            )
            tps_data_in_gaps_contains_zero = np.allclose(
                tps_data_in_gaps, np.zeros(tps_data_in_gaps.size)
            )
            self.assertTrue(tps_data_in_gaps_contains_zero)

    def test_data_has_gap_to_the_left_data_different_day_after_gap(self):
        """
        Test that gaps in the data are skipped in TPS calculation by checking
        that the elements in the TPS array corresponding to the gaps are
        0.

        Test the case where there are gaps to the left of the data and the
        traces directly next to the gaps are in different days.
        """
        # The setup portion of this test suite only create traces in the first
        # positive day in epoch time. So, in order to guarantee there is a gap
        # in the TPS array, we skip the second positive day. The start of the
        # third positive day in epoch time is 172800, so we want a trace that
        # starts after 172801.
        trace_start_time = 173100
        self.add_trace(trace_start_time)

        with self.subTest('test_end_time_same_day_as_second_to_last_trace'):
            # Subject to change after Issue #37 is fixed
            self.tps_processor.end_time = 125000
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            with self.assertRaises(IndexError):
                self.tps_processor.run()

        with self.subTest('test_end_time_cover_all_traces'):
            self.tps_processor.end_time = trace_start_time + 50
            self.tps_processor.each_day_5_mins_list = get_each_day_5_min_list(
                self.tps_processor.start_time,
                self.tps_processor.end_time)
            self.tps_processor.run()
            self.assertEqual(len(self.channel_data['tps_data']), 3)
            tps_gap_day_2 = slice(45, None)
            tps_gap_day_3 = slice(4, None)
            tps_data_in_gaps = np.hstack(
                (
                    self.channel_data['tps_data'][1][tps_gap_day_2],
                    self.channel_data['tps_data'][2][tps_gap_day_3]
                )
            )
            tps_data_in_gaps_contains_zero = np.allclose(
                tps_data_in_gaps, np.zeros(tps_data_in_gaps.size)
            )
            self.assertTrue(tps_data_in_gaps_contains_zero)
